# coding: utf-8

"""
helper routines for testing round trip of commented YAML data
"""
import sys
import textwrap
import io
from pathlib import Path

unset = object()


def dedent(data):
    try:
        position_of_first_newline = data.index('\n')
        for idx in range(position_of_first_newline):
            if not data[idx].isspace():
                raise ValueError
    except ValueError:
        pass
    else:
        data = data[position_of_first_newline + 1 :]
    return textwrap.dedent(data)


def round_trip_load(inp, preserve_quotes=None, version=None):
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    dinp = dedent(inp)
    yaml = ruamel.yaml.YAML()
    yaml.preserve_quotes = preserve_quotes
    yaml.version = version
    return yaml.load(dinp)


def round_trip_load_all(inp, preserve_quotes=None, version=None):
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    dinp = dedent(inp)
    yaml = ruamel.yaml.YAML()
    yaml.preserve_quotes = preserve_quotes
    yaml.version = version
    return yaml.load_all(dinp)


def round_trip_dump(
    data,
    stream=None,  # *,
    indent=None,
    block_seq_indent=None,
    default_flow_style=unset,
    top_level_colon_align=None,
    prefix_colon=None,
    explicit_start=None,
    explicit_end=None,
    version=None,
    allow_unicode=True,
):
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML()
    yaml.indent(mapping=indent, sequence=indent, offset=block_seq_indent)
    if default_flow_style is not unset:
        yaml.default_flow_style = default_flow_style
    yaml.top_level_colon_align = top_level_colon_align
    yaml.prefix_colon = prefix_colon
    yaml.explicit_start = explicit_start
    yaml.explicit_end = explicit_end
    yaml.version = version
    yaml.allow_unicode = allow_unicode
    if stream is not None:
        yaml.dump(data, stream=stream)
        return
    buf = io.StringIO()
    yaml.dump(data, stream=buf)
    return buf.getvalue()


def round_trip_dump_all(
    data,
    stream=None,  # *,
    indent=None,
    block_seq_indent=None,
    default_flow_style=unset,
    top_level_colon_align=None,
    prefix_colon=None,
    explicit_start=None,
    explicit_end=None,
    version=None,
    allow_unicode=None,
):
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML()
    yaml.indent(mapping=indent, sequence=indent, offset=block_seq_indent)
    if default_flow_style is not unset:
        yaml.default_flow_style = default_flow_style
    yaml.top_level_colon_align = top_level_colon_align
    yaml.prefix_colon = prefix_colon
    yaml.explicit_start = explicit_start
    yaml.explicit_end = explicit_end
    yaml.version = version
    yaml.allow_unicode = allow_unicode
    if stream is not None:
        yaml.dump(data, stream=stream)
        return
    buf = io.StringIO()
    yaml.dump_all(data, stream=buf)
    return buf.getvalue()


def diff(inp, outp, file_name='stdin'):
    import difflib

    inl = inp.splitlines(True)  # True for keepends
    outl = outp.splitlines(True)
    diff = difflib.unified_diff(inl, outl, file_name, 'round trip YAML')
    # 2.6 difflib has trailing space on filename lines %-)
    strip_trailing_space = sys.version_info < (2, 7)
    for line in diff:
        if strip_trailing_space and line[:4] in ['--- ', '+++ ']:
            line = line.rstrip() + '\n'
        sys.stdout.write(line)


def round_trip(
    inp,
    outp=None,
    extra=None,
    intermediate=None,
    indent=None,
    block_seq_indent=None,
    top_level_colon_align=None,
    prefix_colon=None,
    preserve_quotes=None,
    explicit_start=None,
    explicit_end=None,
    version=None,
    dump_data=None,
):
    """
    inp:    input string to parse
    outp:   expected output (equals input if not specified)
    """
    if outp is None:
        outp = inp
    doutp = dedent(outp)
    if extra is not None:
        doutp += extra
    data = round_trip_load(inp, preserve_quotes=preserve_quotes)
    if dump_data:
        print('data', data)
    if intermediate is not None:
        if isinstance(intermediate, dict):
            for k, v in intermediate.items():
                if data[k] != v:
                    print('{0!r} <> {1!r}'.format(data[k], v))
                    raise ValueError
    res = round_trip_dump(
        data,
        indent=indent,
        block_seq_indent=block_seq_indent,
        top_level_colon_align=top_level_colon_align,
        prefix_colon=prefix_colon,
        explicit_start=explicit_start,
        explicit_end=explicit_end,
        version=version,
    )
    if res != doutp:
        diff(doutp, res, 'input string')
    print('\nroundtrip data:\n', res, sep="")
    assert res == doutp
    res = round_trip_dump(
        data,
        indent=indent,
        block_seq_indent=block_seq_indent,
        top_level_colon_align=top_level_colon_align,
        prefix_colon=prefix_colon,
        explicit_start=explicit_start,
        explicit_end=explicit_end,
        version=version,
    )
    print('roundtrip second round data:\n', res, sep="")
    assert res == doutp
    return data


def na_round_trip(
    inp,
    outp=None,
    extra=None,
    intermediate=None,
    indent=None,
    top_level_colon_align=None,
    prefix_colon=None,
    preserve_quotes=None,
    explicit_start=None,
    explicit_end=None,
    version=None,
    dump_data=None,
):
    """
    inp:    input string to parse
    outp:   expected output (equals input if not specified)
    """
    inp = dedent(inp)
    if outp is None:
        outp = inp
    if version is not None:
        version = version
    doutp = dedent(outp)
    if extra is not None:
        doutp += extra
    yaml = YAML()
    yaml.preserve_quotes = preserve_quotes
    yaml.scalar_after_indicator = False  # newline after every directives end
    data = yaml.load(inp)
    if dump_data:
        print('data', data)
    if intermediate is not None:
        if isinstance(intermediate, dict):
            for k, v in intermediate.items():
                if data[k] != v:
                    print('{0!r} <> {1!r}'.format(data[k], v))
                    raise ValueError
    yaml.indent = indent
    yaml.top_level_colon_align = top_level_colon_align
    yaml.prefix_colon = prefix_colon
    yaml.explicit_start = explicit_start
    yaml.explicit_end = explicit_end
    res = yaml.dump(data, compare=doutp)
    return res


def YAML(**kw):
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    class MyYAML(ruamel.yaml.YAML):
        """auto dedent string parameters on load"""

        def load(self, stream):
            if isinstance(stream, str):
                if stream and stream[0] == '\n':
                    stream = stream[1:]
                stream = textwrap.dedent(stream)
            return ruamel.yaml.YAML.load(self, stream)

        def load_all(self, stream):
            if isinstance(stream, str):
                if stream and stream[0] == '\n':
                    stream = stream[1:]
                stream = textwrap.dedent(stream)
            for d in ruamel.yaml.YAML.load_all(self, stream):
                yield d

        def dump(self, data, **kw):
            from ruamel.yaml.compat import StringIO, BytesIO  # NOQA

            assert ('stream' in kw) ^ ('compare' in kw)
            if 'stream' in kw:
                return ruamel.yaml.YAML.dump(data, **kw)
            lkw = kw.copy()
            expected = textwrap.dedent(lkw.pop('compare'))
            unordered_lines = lkw.pop('unordered_lines', False)
            if expected and expected[0] == '\n':
                expected = expected[1:]
            lkw['stream'] = st = StringIO()
            ruamel.yaml.YAML.dump(self, data, **lkw)
            res = st.getvalue()
            print(res)
            if unordered_lines:
                res = sorted(res.splitlines())
                expected = sorted(expected.splitlines())
            assert res == expected

        def round_trip(self, stream, **kw):
            from ruamel.yaml.compat import StringIO, BytesIO  # NOQA

            assert isinstance(stream, str)
            lkw = kw.copy()
            if stream and stream[0] == '\n':
                stream = stream[1:]
            stream = textwrap.dedent(stream)
            data = ruamel.yaml.YAML.load(self, stream)
            outp = lkw.pop('outp', stream)
            lkw['stream'] = st = StringIO()
            ruamel.yaml.YAML.dump(self, data, **lkw)
            res = st.getvalue()
            if res != outp:
                diff(outp, res, 'input string')
            assert res == outp

        def round_trip_all(self, stream, **kw):
            from ruamel.yaml.compat import StringIO, BytesIO  # NOQA

            assert isinstance(stream, str)
            lkw = kw.copy()
            if stream and stream[0] == '\n':
                stream = stream[1:]
            stream = textwrap.dedent(stream)
            data = list(ruamel.yaml.YAML.load_all(self, stream))
            outp = lkw.pop('outp', stream)
            lkw['stream'] = st = StringIO()
            ruamel.yaml.YAML.dump_all(self, data, **lkw)
            res = st.getvalue()
            if res != outp:
                diff(outp, res, 'input string')
            assert res == outp

    return MyYAML(**kw)


def save_and_run(program, base_dir=None, output=None, file_name=None, optimized=False):
    """
    safe and run a python program, thereby circumventing any restrictions on module level
    imports
    """
    from subprocess import check_output, STDOUT, CalledProcessError

    if not hasattr(base_dir, 'hash'):
        base_dir = Path(str(base_dir))
    if file_name is None:
        file_name = 'safe_and_run_tmp.py'
    file_name = base_dir / file_name
    file_name.write_text(dedent(program))

    try:
        cmd = [sys.executable, '-Wd']
        if optimized:
            cmd.append('-O')
        cmd.append(str(file_name))
        print('running:', *cmd)
        # 3.5 needs strings
        res = check_output(cmd, stderr=STDOUT, universal_newlines=True, cwd=str(base_dir))
        if output is not None:
            if '__pypy__' in sys.builtin_module_names:
                res = res.splitlines(True)
                res = [line for line in res if 'no version info' not in line]
                res = ''.join(res)
            print('result:  ', res, end='')
            print('expected:', output, end='')
            assert res == output
    except CalledProcessError as exception:
        print("##### Running '{} {}' FAILED #####".format(sys.executable, file_name))
        print(exception.output)
        return exception.returncode
    return 0



# coding: utf-8

import pytest  # NOQA

from roundtrip import round_trip, round_trip_load_all, round_trip_dump_all


class TestDocument:
    def test_single_doc_begin_end(self):
        inp = """\
        ---
        - a
        - b
        ...
        """
        round_trip(inp, explicit_start=True, explicit_end=True)

    def test_multi_doc_begin_end(self):
        inp = """\
        ---
        - a
        ...
        ---
        - b
        ...
        """
        docs = list(round_trip_load_all(inp))
        assert docs == [['a'], ['b']]
        out = round_trip_dump_all(docs, explicit_start=True, explicit_end=True)
        assert out == '---\n- a\n...\n---\n- b\n...\n'

    def test_multi_doc_no_start(self):
        inp = """\
        - a
        ...
        ---
        - b
        ...
        """
        docs = list(round_trip_load_all(inp))
        assert docs == [['a'], ['b']]

    def test_multi_doc_no_end(self):
        inp = """\
        - a
        ---
        - b
        """
        docs = list(round_trip_load_all(inp))
        assert docs == [['a'], ['b']]

    def test_multi_doc_ends_only(self):
        # this is ok in 1.2
        inp = """\
        - a
        ...
        - b
        ...
        """
        docs = list(round_trip_load_all(inp, version=(1, 2)))
        assert docs == [['a'], ['b']]

    def test_multi_doc_ends_only_1_1(self):
        from ruamel import yaml

        # this is not ok in 1.1
        with pytest.raises(yaml.parser.ParserError):
            inp = """\
            - a
            ...
            - b
            ...
            """
            docs = list(round_trip_load_all(inp, version=(1, 1)))
            assert docs == [['a'], ['b']]  # not True, but not reached



# coding: utf-8

"""
testing of anchors and the aliases referring to them
"""

import sys
import textwrap
import pytest
from pathlib import Path


class TestNewAPI:
    def test_duplicate_keys_00(self):
        from  import YAML
        from .constructor import DuplicateKeyError

        yaml = YAML()
        with pytest.raises(DuplicateKeyError):
            yaml.load('{a: 1, a: 2}')

    def test_duplicate_keys_01(self):
        from  import YAML
        from .constructor import DuplicateKeyError

        yaml = YAML(typ='safe', pure=True)
        with pytest.raises(DuplicateKeyError):
            yaml.load('{a: 1, a: 2}')

    def test_duplicate_keys_02(self):
        from  import YAML
        from .constructor import DuplicateKeyError

        yaml = YAML(typ='safe')
        with pytest.raises(DuplicateKeyError):
            yaml.load('{a: 1, a: 2}')

    def test_issue_135(self):
        # reported by Andrzej Ostrowski
        from  import YAML

        data = {'a': 1, 'b': 2}
        yaml = YAML(typ='safe')
        # originally on 2.7: with pytest.raises(TypeError):
        yaml.dump(data, sys.stdout)

    def test_issue_135_temporary_workaround(self):
        # never raised error
        from  import YAML

        data = {'a': 1, 'b': 2}
        yaml = YAML(typ='safe', pure=True)
        yaml.dump(data, sys.stdout)


class TestWrite:
    def test_dump_path(self, tmpdir):
        from  import YAML

        fn = Path(str(tmpdir)) / 'test.yaml'
        yaml = YAML()
        data = yaml.map()
        data['a'] = 1
        data['b'] = 2
        yaml.dump(data, fn)
        assert fn.read_text() == 'a: 1\nb: 2\n'

    def test_dump_file(self, tmpdir):
        from  import YAML

        fn = Path(str(tmpdir)) / 'test.yaml'
        yaml = YAML()
        data = yaml.map()
        data['a'] = 1
        data['b'] = 2
        with open(str(fn), 'w') as fp:
            yaml.dump(data, fp)
        assert fn.read_text() == 'a: 1\nb: 2\n'

    def test_dump_missing_stream(self):
        from  import YAML

        yaml = YAML()
        data = yaml.map()
        data['a'] = 1
        data['b'] = 2
        with pytest.raises(TypeError):
            yaml.dump(data)

    def test_dump_too_many_args(self, tmpdir):
        from  import YAML

        fn = Path(str(tmpdir)) / 'test.yaml'
        yaml = YAML()
        data = yaml.map()
        data['a'] = 1
        data['b'] = 2
        with pytest.raises(TypeError):
            yaml.dump(data, fn, True)

    def test_transform(self, tmpdir):
        from  import YAML

        def tr(s):
            return s.replace(' ', '  ')

        fn = Path(str(tmpdir)) / 'test.yaml'
        yaml = YAML()
        data = yaml.map()
        data['a'] = 1
        data['b'] = 2
        yaml.dump(data, fn, transform=tr)
        assert fn.read_text() == 'a:  1\nb:  2\n'

    def test_print(self, capsys):
        from  import YAML

        yaml = YAML()
        data = yaml.map()
        data['a'] = 1
        data['b'] = 2
        yaml.dump(data, sys.stdout)
        out, err = capsys.readouterr()
        assert out == 'a: 1\nb: 2\n'


class TestRead:
    def test_multi_load(self):
        # make sure reader, scanner, parser get reset
        from  import YAML

        yaml = YAML()
        yaml.load('a: 1')
        yaml.load('a: 1')  # did not work in 0.15.4

    def test_parse(self):
        # ensure `parse` method is functional and can parse "unsafe" yaml
        from  import YAML
        from .constructor import ConstructorError

        yaml = YAML(typ='safe')
        s = '- !User0 {age: 18, name: Anthon}'
        # should fail to load
        with pytest.raises(ConstructorError):
            yaml.load(s)
        # should parse fine
        yaml = YAML(typ='safe')
        for _ in yaml.parse(s):
            pass


class TestLoadAll:
    def test_multi_document_load(self, tmpdir):
        """this went wrong on 3.7 because of StopIteration, PR 37 and Issue 211"""
        from  import YAML

        fn = Path(str(tmpdir)) / 'test.yaml'
        fn.write_text(
            textwrap.dedent("""\
            ---
            - a
            ---
            - b
            ...
            """)
        )
        yaml = YAML()
        assert list(yaml.load_all(fn)) == [['a'], ['b']]


class TestDuplSet:
    def test_dupl_set_00(self):
        # round-trip-loader should except
        from  import YAML
        from .constructor import DuplicateKeyError

        yaml = YAML()
        with pytest.raises(DuplicateKeyError):
            yaml.load(
                textwrap.dedent("""\
                !!set
                ? a
                ? b
                ? c
                ? a
                """)
            )


class TestDumpLoadUnicode:
    # test triggered by SamH on stackoverflow (https://stackoverflow.com/q/45281596/1307905)
    # and answer by randomir (https://stackoverflow.com/a/45281922/1307905)
    def test_write_unicode(self, tmpdir):
        from  import YAML

        yaml = YAML()
        text_dict = {'text': 'HELLO_WORLD©'}
        file_name = str(tmpdir) + '/tstFile.yaml'
        yaml.dump(text_dict, open(file_name, 'w'))
        assert open(file_name, 'rb').read().decode('utf-8') == 'text: HELLO_WORLD©\n'

    def test_read_unicode(self, tmpdir):
        from  import YAML

        yaml = YAML()
        file_name = str(tmpdir) + '/tstFile.yaml'
        with open(file_name, 'wb') as fp:
            fp.write('text: HELLO_WORLD©\n'.encode('utf-8'))
        text_dict = yaml.load(open(file_name, 'r'))
        assert text_dict['text'] == 'HELLO_WORLD©'


class TestFlowStyle:
    def test_flow_style(self, capsys):
        # https://stackoverflow.com/questions/45791712/
        from  import YAML

        yaml = YAML()
        yaml.default_flow_style = None
        data = yaml.map()
        data['b'] = 1
        data['a'] = [[1, 2], [3, 4]]
        yaml.dump(data, sys.stdout)
        out, err = capsys.readouterr()
        assert out == 'b: 1\na:\n- [1, 2]\n- [3, 4]\n'


class TestOldAPI:
    @pytest.mark.skipif(sys.version_info >= (3, 0), reason='ok on Py3')
    def test_duplicate_keys_02(self):
        # Issue 165 unicode keys in error/warning
        from  import safe_load
        from .constructor import DuplicateKeyError

        with pytest.raises(DuplicateKeyError):
            safe_load('type: Doméstica\ntype: International')



# coding: utf-8

import sys
import os
import pytest  # NOQA

sys.path.insert(0, os.path.dirname(__file__) + '/lib')

import warnings  # NOQA


args = []


def test_data():
    import test_appliance  # NOQA

    warnings.simplefilter('ignore', PendingDeprecationWarning)
    collections = []
    import test_yaml

    collections.append(test_yaml)
    test_appliance.run(collections, args)

# @pytest.mark.skipif(not ruamel.yaml.__with_libyaml__,
#                     reason="no libyaml")


def test_data_ext():
    collections = []
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA
    import test_appliance  # NOQA

    warnings.simplefilter('ignore', ruamel.yaml.error.UnsafeLoaderWarning)
    warnings.simplefilter('ignore', PendingDeprecationWarning)
    if ruamel.yaml.__with_libyaml__:
        import test_yaml_ext

        collections.append(test_yaml_ext)
        test_appliance.run(collections, args)



# coding: utf-8

try:
    import numpy
except:  # NOQA
    numpy = None


def Xtest_numpy():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")

    if numpy is None:
        return
    data = numpy.arange(10)
    print('data', type(data), data)

    yaml_str = ruamel.yaml.dump(data)
    datb = ruamel.yaml.load(yaml_str)
    print('datb', type(datb), datb)

    print('\nYAML', yaml_str)
    assert data == datb



import ez_yaml

value = ez_yaml.to_object(file_path="./tests/info.yaml")
print(f'''value = {value}''')

value = ez_yaml.to_object(file_path="./tests/info.yaml", load_nested_yaml=True)
print(f'''value = {value}''')


# coding: utf-8

"""
various test cases for YAML files
"""

import sys
import io
import pytest  # NOQA
import platform

from roundtrip import round_trip, dedent, round_trip_load, round_trip_dump  # NOQA


class TestYAML:
    def test_backslash(self):
        round_trip("""
        handlers:
          static_files: applications/\\1/static/\\2
        """)

    def test_omap_out(self):
        # ordereddict mapped to !!omap
        from ruamel.yaml.compat import ordereddict
        exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

        x = ordereddict([('a', 1), ('b', 2)])
        res = round_trip_dump(x, default_flow_style=False)
        assert res == dedent("""
        !!omap
        - a: 1
        - b: 2
        """)

    def test_omap_roundtrip(self):
        round_trip("""
        !!omap
        - a: 1
        - b: 2
        - c: 3
        - d: 4
        """)

    @pytest.mark.skipif(sys.version_info < (2, 7), reason='collections not available')
    def test_dump_collections_ordereddict(self):
        from collections import OrderedDict
        exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

        # OrderedDict mapped to !!omap
        x = OrderedDict([('a', 1), ('b', 2)])
        res = round_trip_dump(x, default_flow_style=False)
        assert res == dedent("""
        !!omap
        - a: 1
        - b: 2
        """)

    @pytest.mark.skipif(
        sys.version_info >= (3, 0) or platform.python_implementation() != 'CPython',
        reason='ruamel.yaml not available',
    )
    def test_dump_ruamel_ordereddict(self):
        from ruamel.ordereddict import ordereddict
        exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

        # OrderedDict mapped to !!omap
        x = ordereddict([('a', 1), ('b', 2)])
        res = round_trip_dump(x, default_flow_style=False)
        assert res == dedent("""
        !!omap
        - a: 1
        - b: 2
        """)

    def test_CommentedSet(self):
        from ruamel.yaml.constructor import CommentedSet

        s = CommentedSet(['a', 'b', 'c'])
        s.remove('b')
        s.add('d')
        assert s == CommentedSet(['a', 'c', 'd'])
        s.add('e')
        s.add('f')
        s.remove('e')
        assert s == CommentedSet(['a', 'c', 'd', 'f'])

    def test_set_out(self):
        # preferable would be the shorter format without the ': null'
        exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

        x = set(['a', 'b', 'c'])
        # cannot use round_trip_dump, it doesn't show null in block style
        buf = io.StringIO()
        yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
        yaml.default_flow_style = False
        yaml.dump(x, buf)
        assert buf.getvalue() == dedent("""
        !!set
        a: null
        b: null
        c: null
        """)

    # ordering is not preserved in a set
    def test_set_compact(self):
        # this format is read and also should be written by default
        round_trip("""
        !!set
        ? a
        ? b
        ? c
        """)

    def test_blank_line_after_comment(self):
        round_trip("""
        # Comment with spaces after it.


        a: 1
        """)

    def test_blank_line_between_seq_items(self):
        round_trip("""
        # Seq with empty lines in between items.
        b:
        - bar


        - baz
        """)

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_blank_line_after_literal_chip(self):
        s = """
        c:
        - |
          This item
          has a blank line
          following it.

        - |
          To visually separate it from this item.

          This item contains a blank line.


        """
        d = round_trip_load(dedent(s))
        print(d)
        round_trip(s)
        assert d['c'][0].split('it.')[1] == '\n'
        assert d['c'][1].split('line.')[1] == '\n'

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_blank_line_after_literal_keep(self):
        """ have to insert an eof marker in YAML to test this"""
        s = """
        c:
        - |+
          This item
          has a blank line
          following it.

        - |+
          To visually separate it from this item.

          This item contains a blank line.


        ...
        """
        d = round_trip_load(dedent(s))
        print(d)
        round_trip(s)
        assert d['c'][0].split('it.')[1] == '\n\n'
        assert d['c'][1].split('line.')[1] == '\n\n\n'

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_blank_line_after_literal_strip(self):
        s = """
        c:
        - |-
          This item
          has a blank line
          following it.

        - |-
          To visually separate it from this item.

          This item contains a blank line.


        """
        d = round_trip_load(dedent(s))
        print(d)
        round_trip(s)
        assert d['c'][0].split('it.')[1] == ""
        assert d['c'][1].split('line.')[1] == ""

    def test_load_all_perserve_quotes(self):
        exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        s = dedent("""\
        a: 'hello'
        ---
        b: "goodbye"
        """)
        data = []
        for x in yaml.load_all(s):
            data.append(x)
        buf = ruamel.yaml.compat.StringIO()
        yaml.dump_all(data, buf)
        out = buf.getvalue()
        print(type(data[0]['a']), data[0]['a'])
        # out = ruamel.yaml.round_trip_dump_all(data)
        print(out)
        assert out == s




exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")


class AnInstance:
    def __init__(self, foo, bar):
        self.foo = foo
        self.bar = bar

    def __repr__(self):
        try:
            return '%s(foo=%r, bar=%r)' % (self.__class__.__name__, self.foo, self.bar)
        except RuntimeError:
            return '%s(foo=..., bar=...)' % self.__class__.__name__


class AnInstanceWithState(AnInstance):
    def __getstate__(self):
        return {'attributes': [self.foo, self.bar]}

    def __setstate__(self, state):
        self.foo, self.bar = state['attributes']


def test_recursive(recursive_filename, verbose=False):
    yaml = ruamel.yaml.YAML(typ='safe', pure=True)
    context = globals().copy()
    with open(recursive_filename, 'rb') as fp0:
        exec(fp0.read(), context)
    value1 = context['value']
    output1 = None
    value2 = None
    output2 = None
    try:
        buf = ruamel.yaml.compat.StringIO()
        output1 = yaml.dump(value1, buf)
        yaml.load(output1)
        value2 = buf.getvalue()
        buf = ruamel.yaml.compat.StringIO()
        yaml.dump(value2, buf)
        output2 = buf.getvalue()
        assert output1 == output2, (output1, output2)
    finally:
        if verbose:
            print('VALUE1:', value1)
            print('VALUE2:', value2)
            print('OUTPUT1:')
            print(output1)
            print('OUTPUT2:')
            print(output2)


test_recursive.unittest = ['.recursive']

if __name__ == '__main__':
    import test_appliance

    test_appliance.run(globals())



from os import remove, getcwd, makedirs, listdir, rename, rmdir, system
from os.path import isabs, isfile, isdir, join, dirname, basename, exists, splitext, relpath, join, dirname
from pathlib import Path
import os
import shutil
import sys
import warnings # TODO: convert serveral prints to warnings
from hashlib import md5 

# 
# helpers
# 
def consistent_hash(value):
    if isinstance(value, bytes):
        return md5(value).hexdigest()
    
    if isinstance(value, str):
        return md5(("@"+value).encode('utf-8')).hexdigest()
    
    if isinstance(value, (bool, int, float, type(None))):
        return md5(str(value).encode('utf-8')).hexdigest()
        
    else:
        return md5(pickle.dumps(value, protocol=4)).hexdigest()

def make_absolute_path(to, coming_from=None):
    # if coming from cwd, its easy
    if coming_from is None:
        return os.path.abspath(to)
    
    # source needs to be absolute
    coming_from_absolute = os.path.abspath(coming_from)
    # if other path is  absolute, make it relative to coming_from
    relative_path = to
    if os.path.isabs(to):
        relative_path = os.path.relpath(to, coming_from_absolute)
    return os.path.join(coming_from_absolute, relative_path)

def make_relative_path(*, to, coming_from=None):
    if coming_from is None:
        coming_from = get_cwd()
    return os.path.relpath(to, coming_from)

def path_pieces(path):
    """
    example:
        *folders, file_name, file_extension = path_pieces("/this/is/a/filepath.txt")
    """
    folders = []
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        else:
            if path != "":
                folders.append(path)

            break
    folders.reverse()
    *folders, file = folders
    filename, file_extension = os.path.splitext(file)
    return [ *folders, filename, file_extension ]

def remove(path):
    if not Path(path).is_symlink() and os.path.isdir(path):
        shutil.rmtree(path)
    else:
        try:
            os.remove(path)
        except:
            pass

def final_target_of(path):
    # resolve symlinks
    if os.path.islink(path):
        have_seen = set()
        while os.path.islink(path):
            path = os.readlink(path)
            if path in have_seen:
                return None # circular broken link
            have_seen.add(path)
    return path
    
# 
# globals
# 
this_file = make_absolute_path(__file__)
this_folder = dirname(this_file)
settings_path = join(this_folder, '..', 'settings.json')

# 
# find closest import path
#
*folders_to_this, _, _ = path_pieces(this_file)
best_match_amount = 0
best_import_zone_match = None
import sys
for each_import_path in sys.path:
    each_import_path = make_absolute_path(each_import_path)
    if not isdir(each_import_path):
        continue
    *folders_of_import_path, name, extension = path_pieces(each_import_path)
    folders_of_import_path = [ *folders_of_import_path, name+extension]
    matches = 0
    for folder_to_this, folder_to_some_import_zone in zip(folders_to_this, folders_of_import_path):
        if folder_to_this != folder_to_some_import_zone:
            break
        else:
            matches += 1
    
    if matches > best_match_amount:
        best_import_zone_match = each_import_path
        best_match_amount = matches

# shouldn't ever happen
if best_import_zone_match is None:
    raise Exception(f"""Couldn't find a path to {this_file} from any of {sys.path}""")

# 
# import json mapping
# 
import json
from os.path import join
# ensure it exists
if not isfile(settings_path):
    with open(settings_path, 'w') as the_file:
        the_file.write(str("{}"))
with open(settings_path, 'r') as in_file:
    settings = json.load(in_file)
    if not isinstance(settings, dict):
        raise Exception(f"""\n\n\nThis file is corrupt (it should be a JSON object):{settings_path}""")

# ensure that pure_python_packages exists
if not isinstance(settings.get("pure_python_packages", None), dict):
    settings["pure_python_packages"] = {}
dependency_mapping = settings["pure_python_packages"]

# 
# calculate paths
# 
from random import random
counter = 0
import_strings = []
for dependency_name, dependency_info in dependency_mapping.items():
    counter += 1
    if dependency_name.startswith("__"):
        raise Exception(f"""dependency names cannot start with "__", but this one does: {dependency_name}. This source of that name is in: {settings_path}""")
    if not dependency_name.isidentifier():
        raise Exception(f"""dependency names must be an identifier ("blah".isidentifier() in python), but this one is not: {dependency_name}. This source of that name is in: {settings_path}""")
    
    target_path = join(this_folder, dependency_info["path"])
    relative_target_path = make_relative_path(to=target_path, coming_from=best_import_zone_match)
    # ensure the parent folder
    *path_parts, _, _ = path_pieces(join(relative_target_path, "_"))
    eval_part = dependency_info.get("eval", dependency_name)
    unique_name = f"{dependency_name}_{random()}_{counter}".replace(".","")
    target_folder_for_import = join(this_folder, dependency_name)
    if not Path(target_folder_for_import).is_symlink() or final_target_of(target_folder_for_import) != dependency_info["path"]:
        # clear the way
        remove(target_folder_for_import)
        # symlink the folder
        Path(target_folder_for_import).symlink_to(dependency_info["path"])

# import the paths
__all__ = []
for dependency_name, dependency_info in dependency_mapping.items():
    # this will register it with python and convert it to a proper module with a unique path (important for pickling things)
    try:
        exec(f"""from .{dependency_name} import __file__ as _""")
        __all__.append(dependency_name)
    except ImportError as error:
        if f"{error}" == "cannot import name '__file__'":
            # this means top level folder isn't a module or doesnt have a __init__.py
            # some modules simply are like this
            pass
        else:
            raise error

# replacements:
    # this '''
    #     from ruamel\.yaml\.
    # '''
    # replaced by '''
    #     from \.
    # '''
    
    # this '''
    #     ^( *)import ruamel(.*?)(?=#|$)
    # '''
    # had to be manually refactored



if __name__ == '__main__':
    import sys
    import os
    import distutils.util

    build_lib = 'build/lib'
    build_lib_ext = os.path.join(
        'build', 'lib.%s-%s' % (distutils.util.get_platform(), sys.version[0:3])
    )
    sys.path.insert(0, build_lib)
    sys.path.insert(0, build_lib_ext)
    import test_yaml
    import test_appliance

    test_appliance.run(test_yaml)




exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")
import pprint

import datetime

try:
    set
except NameError:
    from sets import Set as set  # NOQA
exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml.tokens;ruamel = {".".join(__name__.split(".")[:-2])}""")


def execute(code):
    global value
    exec(code)
    return value


def _make_objects():
    global MyLoader, MyDumper, MyTestClass1, MyTestClass2, MyTestClass3, YAMLobject1, YAMLobject2, AnObject, AnInstance, AState, ACustomState, InitArgs, InitArgsWithState, NewArgs, NewArgsWithState, Reduce, ReduceWithState, MyInt, MyList, MyDict, FixedOffset, today, execute

    class MyLoader(ruamel.yaml.Loader):
        pass

    class MyDumper(ruamel.yaml.Dumper):
        pass

    class MyTestClass1:
        def __init__(self, x, y=0, z=0):
            self.x = x
            self.y = y
            self.z = z

        def __eq__(self, other):
            if isinstance(other, MyTestClass1):
                return self.__class__, self.__dict__ == other.__class__, other.__dict__
            else:
                return False

    def construct1(constructor, node):
        mapping = constructor.construct_mapping(node)
        return MyTestClass1(**mapping)

    def represent1(representer, native):
        return representer.represent_mapping('!tag1', native.__dict__)

    ruamel.yaml.add_constructor('!tag1', construct1, Loader=MyLoader)
    ruamel.yaml.add_representer(MyTestClass1, represent1, Dumper=MyDumper)

    class MyTestClass2(MyTestClass1, ruamel.yaml.YAMLObject):
        ruamel.yaml.loader = MyLoader
        ruamel.yaml.dumper = MyDumper
        ruamel.yaml.tag = '!tag2'

        def from_yaml(cls, constructor, node):
            x = constructor.construct_yaml_int(node)
            return cls(x=x)

        from_yaml = classmethod(from_yaml)

        def to_yaml(cls, representer, native):
            return representer.represent_scalar(cls.yaml_tag, str(native.x))

        to_yaml = classmethod(to_yaml)

    class MyTestClass3(MyTestClass2):
        ruamel.yaml.tag = '!tag3'

        def from_yaml(cls, constructor, node):
            mapping = constructor.construct_mapping(node)
            if '=' in mapping:
                x = mapping['=']
                del mapping['=']
                mapping['x'] = x
            return cls(**mapping)

        from_yaml = classmethod(from_yaml)

        def to_yaml(cls, representer, native):
            return representer.represent_mapping(cls.yaml_tag, native.__dict__)

        to_yaml = classmethod(to_yaml)

    class YAMLobject1(ruamel.yaml.YAMLObject):
        ruamel.yaml.loader = MyLoader
        ruamel.yaml.dumper = MyDumper
        ruamel.yaml.tag = '!foo'

        def __init__(self, my_parameter=None, my_another_parameter=None):
            self.my_parameter = my_parameter
            self.my_another_parameter = my_another_parameter

        def __eq__(self, other):
            if isinstance(other, YAMLobject1):
                return self.__class__, self.__dict__ == other.__class__, other.__dict__
            else:
                return False

    class YAMLobject2(ruamel.yaml.YAMLObject):
        ruamel.yaml.loader = MyLoader
        ruamel.yaml.dumper = MyDumper
        ruamel.yaml.tag = '!bar'

        def __init__(self, foo=1, bar=2, baz=3):
            self.foo = foo
            self.bar = bar
            self.baz = baz

        def __getstate__(self):
            return {1: self.foo, 2: self.bar, 3: self.baz}

        def __setstate__(self, state):
            self.foo = state[1]
            self.bar = state[2]
            self.baz = state[3]

        def __eq__(self, other):
            if isinstance(other, YAMLobject2):
                return self.__class__, self.__dict__ == other.__class__, other.__dict__
            else:
                return False

    class AnObject(object):
        def __new__(cls, foo=None, bar=None, baz=None):
            self = object.__new__(cls)
            self.foo = foo
            self.bar = bar
            self.baz = baz
            return self

        def __cmp__(self, other):
            return cmp(
                (type(self), self.foo, self.bar, self.baz),  # NOQA
                (type(other), other.foo, other.bar, other.baz),
            )

        def __eq__(self, other):
            return type(self) is type(other) and (self.foo, self.bar, self.baz) == (
                other.foo,
                other.bar,
                other.baz,
            )

    class AnInstance:
        def __init__(self, foo=None, bar=None, baz=None):
            self.foo = foo
            self.bar = bar
            self.baz = baz

        def __cmp__(self, other):
            return cmp(
                (type(self), self.foo, self.bar, self.baz),  # NOQA
                (type(other), other.foo, other.bar, other.baz),
            )

        def __eq__(self, other):
            return type(self) is type(other) and (self.foo, self.bar, self.baz) == (
                other.foo,
                other.bar,
                other.baz,
            )

    class AState(AnInstance):
        def __getstate__(self):
            return {'_foo': self.foo, '_bar': self.bar, '_baz': self.baz}

        def __setstate__(self, state):
            self.foo = state['_foo']
            self.bar = state['_bar']
            self.baz = state['_baz']

    class ACustomState(AnInstance):
        def __getstate__(self):
            return (self.foo, self.bar, self.baz)

        def __setstate__(self, state):
            self.foo, self.bar, self.baz = state

    # class InitArgs(AnInstance):
    #     def __getinitargs__(self):
    #         return (self.foo, self.bar, self.baz)
    #     def __getstate__(self):
    #         return {}

    # class InitArgsWithState(AnInstance):
    #     def __getinitargs__(self):
    #         return (self.foo, self.bar)
    #     def __getstate__(self):
    #         return self.baz
    #     def __setstate__(self, state):
    #         self.baz = state

    class NewArgs(AnObject):
        def __getnewargs__(self):
            return (self.foo, self.bar, self.baz)

        def __getstate__(self):
            return {}

    class NewArgsWithState(AnObject):
        def __getnewargs__(self):
            return (self.foo, self.bar)

        def __getstate__(self):
            return self.baz

        def __setstate__(self, state):
            self.baz = state

    InitArgs = NewArgs

    InitArgsWithState = NewArgsWithState

    class Reduce(AnObject):
        def __reduce__(self):
            return self.__class__, (self.foo, self.bar, self.baz)

    class ReduceWithState(AnObject):
        def __reduce__(self):
            return self.__class__, (self.foo, self.bar), self.baz

        def __setstate__(self, state):
            self.baz = state

    class MyInt(int):
        def __eq__(self, other):
            return type(self) is type(other) and int(self) == int(other)

    class MyList(list):
        def __init__(self, n=1):
            self.extend([None] * n)

        def __eq__(self, other):
            return type(self) is type(other) and list(self) == list(other)

    class MyDict(dict):
        def __init__(self, n=1):
            for k in range(n):
                self[k] = None

        def __eq__(self, other):
            return type(self) is type(other) and dict(self) == dict(other)

    class FixedOffset(datetime.tzinfo):
        def __init__(self, offset, name):
            self.__offset = datetime.timedelta(minutes=offset)
            self.__name = name

        def utcoffset(self, dt):
            return self.__offset

        def tzname(self, dt):
            return self.__name

        def dst(self, dt):
            return datetime.timedelta(0)

    today = datetime.date.today()


try:
    from ruamel.ordereddict import ordereddict
except ImportError:
    from collections import OrderedDict

    # to get the right name import ... as ordereddict doesn't do that

    class ordereddict(OrderedDict):
        pass


def _load_code(expression):
    return eval(expression, globals())


def _serialize_value(data):
    if isinstance(data, list):
        return '[%s]' % ', '.join(map(_serialize_value, data))
    elif isinstance(data, dict):
        items = []
        for key, value in data.items():
            key = _serialize_value(key)
            value = _serialize_value(value)
            items.append('%s: %s' % (key, value))
        items.sort()
        return '{%s}' % ', '.join(items)
    elif isinstance(data, datetime.datetime):
        return repr(data.utctimetuple())
    elif isinstance(data, float) and data != data:
        return '?'
    else:
        return str(data)


def test_constructor_types(data_filename, code_filename, verbose=False):
    _make_objects()
    native1 = None
    native2 = None
    yaml = ruamel.yaml.YAML(typ='safe', pure=True)
    yaml.loader = MyLoader
    try:
        with open(data_filename, 'rb') as fp0:
            native1 = list(yaml.load_all(fp0))
        if len(native1) == 1:
            native1 = native1[0]
        with open(code_filename, 'rb') as fp0:
            native2 = _load_code(fp0.read())
        try:
            if native1 == native2:
                return
        except TypeError:
            pass
        # print('native1', native1)
        if verbose:
            print('SERIALIZED NATIVE1:')
            print(_serialize_value(native1))
            print('SERIALIZED NATIVE2:')
            print(_serialize_value(native2))
        assert _serialize_value(native1) == _serialize_value(native2), (native1, native2)
    finally:
        if verbose:
            print('NATIVE1:')
            pprint.pprint(native1)
            print('NATIVE2:')
            pprint.pprint(native2)


test_constructor_types.unittest = ['.data', '.code']


def test_roundtrip_data(code_filename, roundtrip_filename, verbose=False):
    _make_objects()
    with open(code_filename, 'rb') as fp0:
        value1 = fp0.read()
    yaml = YAML(typ='safe', pure=True)
    yaml.Loader = MyLoader
    native2 = list(yaml.load_all(value1))
    if len(native2) == 1:
        native2 = native2[0]
    try:
        value2 = ruamel.yaml.dump(
            native2,
            Dumper=MyDumper,
            default_flow_style=False,
            allow_unicode=True,
            encoding='utf-8',
        )
        # value2 += x
        if verbose:
            print('SERIALIZED NATIVE1:')
            print(value1)
            print('SERIALIZED NATIVE2:')
            print(value2)
        assert value1 == value2, (value1, value2)
    finally:
        if verbose:
            print('NATIVE2:')
            pprint.pprint(native2)


test_roundtrip_data.unittest = ['.data', '.roundtrip']


if __name__ == '__main__':
    import sys
    import test_constructor  # NOQA

    sys.modules['test_constructor'] = sys.modules['__main__']
    import test_appliance

    test_appliance.run(globals())



from __future__ import absolute_import
from __future__ import print_function

exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml as yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")


def _compare_events(events1, events2):
    assert len(events1) == len(events2), (events1, events2)
    for event1, event2 in zip(events1, events2):
        assert event1.__class__ == event2.__class__, (event1, event2)
        if isinstance(event1, yaml.NodeEvent):
            assert event1.anchor == event2.anchor, (event1, event2)
        if isinstance(event1, yaml.CollectionStartEvent):
            assert event1.tag == event2.tag, (event1, event2)
        if isinstance(event1, yaml.ScalarEvent):
            if True not in event1.implicit + event2.implicit:
                assert event1.tag == event2.tag, (event1, event2)
            assert event1.value == event2.value, (event1, event2)


def test_emitter_on_data(data_filename, canonical_filename, verbose=False):
    with open(data_filename, 'rb') as fp0:
        events = list(yaml.parse(fp0))
    output = yaml.emit(events)
    if verbose:
        print('OUTPUT:')
        print(output)
    new_events = list(yaml.parse(output))
    _compare_events(events, new_events)


test_emitter_on_data.unittest = ['.data', '.canonical']


def test_emitter_on_canonical(canonical_filename, verbose=False):
    with open(canonical_filename, 'rb') as fp0:
        events = list(yaml.parse(fp0))
    for canonical in [False, True]:
        output = yaml.emit(events, canonical=canonical)
        if verbose:
            print('OUTPUT (canonical=%s):' % canonical)
            print(output)
        new_events = list(yaml.parse(output))
        _compare_events(events, new_events)


test_emitter_on_canonical.unittest = ['.canonical']


def test_emitter_styles(data_filename, canonical_filename, verbose=False):
    for filename in [data_filename, canonical_filename]:
        with open(filename, 'rb') as fp0:
            events = list(yaml.parse(fp0))
        for flow_style in [False, True]:
            for style in ['|', '>', '"', "'", ""]:
                styled_events = []
                for event in events:
                    if isinstance(event, yaml.ScalarEvent):
                        event = yaml.ScalarEvent(
                            event.anchor, event.tag, event.implicit, event.value, style=style
                        )
                    elif isinstance(event, yaml.SequenceStartEvent):
                        event = yaml.SequenceStartEvent(
                            event.anchor, event.tag, event.implicit, flow_style=flow_style
                        )
                    elif isinstance(event, yaml.MappingStartEvent):
                        event = yaml.MappingStartEvent(
                            event.anchor, event.tag, event.implicit, flow_style=flow_style
                        )
                    styled_events.append(event)
                output = yaml.emit(styled_events)
                if verbose:
                    print(
                        'OUTPUT (filename=%r, flow_style=%r, style=%r)'
                        % (filename, flow_style, style)
                    )
                    print(output)
                new_events = list(yaml.parse(output))
                _compare_events(events, new_events)


test_emitter_styles.unittest = ['.data', '.canonical']


class EventsLoader(yaml.Loader):
    def construct_event(self, node):
        if isinstance(node, yaml.ScalarNode):
            mapping = {}
        else:
            mapping = self.construct_mapping(node)
        class_name = str(node.tag[1:]) + 'Event'
        if class_name in [
            'AliasEvent',
            'ScalarEvent',
            'SequenceStartEvent',
            'MappingStartEvent',
        ]:
            mapping.setdefault('anchor', None)
        if class_name in ['ScalarEvent', 'SequenceStartEvent', 'MappingStartEvent']:
            mapping.setdefault('tag', None)
        if class_name in ['SequenceStartEvent', 'MappingStartEvent']:
            mapping.setdefault('implicit', True)
        if class_name == 'ScalarEvent':
            mapping.setdefault('implicit', (False, True))
            mapping.setdefault('value', "")
        value = getattr(yaml, class_name)(**mapping)
        return value


# if Loader is not a composite, add this function
# EventsLoader.add_constructor = yaml.constructor.Constructor.add_constructor


EventsLoader.add_constructor(None, EventsLoader.construct_event)


def test_emitter_events(events_filename, verbose=False):
    with open(events_filename, 'rb') as fp0:
        events = list(yaml.load(fp0, Loader=EventsLoader))
    output = yaml.emit(events)
    if verbose:
        print('OUTPUT:')
        print(output)
    new_events = list(yaml.parse(output))
    _compare_events(events, new_events)


if __name__ == '__main__':
    import test_appliance

    test_appliance.run(globals())



import os

def walk_up_until(file_to_find, start_path=None):
    here = start_path or os.getcwd()
    if not os.path.isabs(here):
        here = os.path.join(os.getcwd(), file_to_find)
    
    while 1:
        check_path = os.path.join(here, file_to_find)
        if os.path.exists(check_path):
            return check_path
        
        # reached the top
        if here == os.path.dirname(here):
            return None
        else:
            # go up a folder
            here = os.path.dirname(here)


# coding: utf-8

import re
import pytest  # NOQA

from roundtrip import dedent, round_trip_dump  # NOQA


# from PyYAML docs
class Dice(tuple):
    def __new__(cls, a, b):
        return tuple.__new__(cls, [a, b])

    def __repr__(self):
        return 'Dice(%s,%s)' % self


def dice_constructor(loader, node):
    value = loader.construct_scalar(node)
    a, b = map(int, value.split('d'))
    return Dice(a, b)


def dice_representer(dumper, data):
    return dumper.represent_scalar('!dice', '{}d{}'.format(*data))


def test_dice_constructor():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
    ruamel.yaml.add_constructor('!dice', dice_constructor)
    data = yaml.load('initial hit points: !dice 8d4')
    assert str(data) == "{'initial hit points': Dice(8,4)}"


def test_dice_constructor_with_loader():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
    ruamel.yaml.add_constructor('!dice', dice_constructor, Loader=ruamel.yaml.Loader)
    data = yaml.load('initial hit points: !dice 8d4')
    assert str(data) == "{'initial hit points': Dice(8,4)}"


def test_dice_representer():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
    yaml.default_flow_style = False
    ruamel.yaml.add_representer(Dice, dice_representer)
    # ruamel.yaml 0.15.8+ no longer forces quotes tagged scalars
    buf = ruamel.yaml.compat.StringIO()
    yaml.dump(dict(gold=Dice(10, 6)), buf)
    assert buf.getvalue() == 'gold: !dice 10d6\n'


def test_dice_implicit_resolver():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
    yaml.default_flow_style = False
    pattern = re.compile(r'^\d+d\d+$')
    ruamel.yaml.add_implicit_resolver('!dice', pattern)
    buf = ruamel.yaml.compat.StringIO()
    yaml.dump(dict(treasure=Dice(10, 20)), buf)
    assert buf.getvalue() == 'treasure: 10d20\n'
    assert yaml.load('damage: 5d10') == dict(damage=Dice(5, 10))


class Obj1(dict):
    def __init__(self, suffix):
        self._suffix = suffix
        self._node = None

    def add_node(self, n):
        self._node = n

    def __repr__(self):
        return 'Obj1(%s->%s)' % (self._suffix, self.items())

    def dump(self):
        return repr(self._node)


class YAMLObj1(object):
    yaml_tag = '!obj:'

    @classmethod
    def from_yaml(cls, loader, suffix, node):
        exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

        obj1 = Obj1(suffix)
        if isinstance(node, ruamel.yaml.MappingNode):
            obj1.add_node(loader.construct_mapping(node))
        else:
            raise NotImplementedError
        return obj1

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(cls.yaml_tag + data._suffix, data.dump())


def test_yaml_obj():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
    ruamel.yaml.add_representer(Obj1, YAMLObj1.to_yaml)
    ruamel.yaml.add_multi_constructor(YAMLObj1.yaml_tag, YAMLObj1.from_yaml)
    x = yaml.load('!obj:x.2\na: 1')
    print(x)
    buf = ruamel.yaml.compat.StringIO()
    yaml.dump(x, buf)
    assert buf.getvalue() == """!obj:x.2 "{'a': 1}"\n"""


def test_yaml_obj_with_loader_and_dumper():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    yaml = ruamel.yaml.YAML(typ='unsafe', pure=True)
    ruamel.yaml.add_representer(Obj1, YAMLObj1.to_yaml, Dumper=ruamel.yaml.Dumper)
    ruamel.yaml.add_multi_constructor(
        YAMLObj1.yaml_tag, YAMLObj1.from_yaml, Loader=ruamel.yaml.Loader
    )
    x = yaml.load('!obj:x.2\na: 1')
    # x = ruamel.yaml.load('!obj:x.2\na: 1')
    print(x)
    buf = ruamel.yaml.compat.StringIO()
    yaml.dump(x, buf)
    assert buf.getvalue() == """!obj:x.2 "{'a': 1}"\n"""


# ToDo use nullege to search add_multi_representer and add_path_resolver
# and add some test code

# Issue 127 reported by Tommy Wang


def test_issue_127():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml  ;ruamel = {".".join(__name__.split(".")[:-2])}""")# NOQA

    class Ref(ruamel.yaml.YAMLObject):
        yaml_constructor = ruamel.yaml.RoundTripConstructor
        yaml_representer = ruamel.yaml.RoundTripRepresenter
        yaml_tag = '!Ref'

        def __init__(self, logical_id):
            self.logical_id = logical_id

        @classmethod
        def from_yaml(cls, loader, node):
            return cls(loader.construct_scalar(node))

        @classmethod
        def to_yaml(cls, dumper, data):
            if isinstance(data.logical_id, ruamel.yaml.scalarstring.ScalarString):
                style = data.logical_id.style  # ruamel.yaml>0.15.8
            else:
                style = None
            return dumper.represent_scalar(cls.yaml_tag, data.logical_id, style=style)

    document = dedent("""\
    AList:
      - !Ref One
      - !Ref 'Two'
      - !Ref
        Two and a half
    BList: [!Ref Three, !Ref "Four"]
    CList:
      - Five Six
      - 'Seven Eight'
    """)
    yaml = ruamel.yaml.YAML()
    yaml.preserve_quotes = True
    yaml.default_flow_style = None
    yaml.indent(sequence=4, offset=2)
    data = yaml.load(document)
    buf = ruamel.yaml.compat.StringIO()
    yaml.dump(data, buf)
    assert buf.getvalue() == document.replace('\n    Two and', ' Two and')



# coding: utf-8

import sys
import platform
import pytest
from textwrap import dedent

NO_CLIB_VER = (3, 10)


@pytest.mark.skipif(
    platform.python_implementation() in ['Jython', 'PyPy'],
    reason='Jython throws RepresenterError'
)
def test_load_cyaml():
    print("???????????????????????", platform.python_implementation())
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")

    if sys.version_info >= NO_CLIB_VER:
        return
    yaml = ruamel.yaml.YAML(typ='safe', pure=False)
    assert ruamel.yaml.__with_libyaml__

    yaml.load('abc: 1')


@pytest.mark.skipif(sys.version_info >= NO_CLIB_VER
                    or platform.python_implementation() in ['Jython', 'PyPy'],
                    reason='no _PyGC_FINALIZED')
def test_dump_cyaml():
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")

    if sys.version_info >= NO_CLIB_VER:
        return
    data = {'a': 1, 'b': 2}
    yaml = ruamel.yaml.YAML(typ='safe', pure=False)
    yaml.default_flow_style = False
    yaml.allow_unicode = True
    buf = ruamel.yaml.compat.StringIO()
    yaml.dump(data, buf)
    assert buf.getvalue() == 'a: 1\nb: 2\n'


@pytest.mark.skipif(
    platform.python_implementation() in ['Jython', 'PyPy'], reason='not avialable'
)
def test_load_cyaml_1_2():
    # issue 155
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")

    if sys.version_info >= NO_CLIB_VER:
        return
    assert ruamel.yaml.__with_libyaml__
    inp = dedent("""\
    %YAML 1.2
    ---
    num_epochs: 70000
    """)
    yaml = ruamel.yaml.YAML(typ='safe')
    yaml.load(inp)


@pytest.mark.skipif(
    platform.python_implementation() in ['Jython', 'PyPy'], reason='not available'
)
def test_dump_cyaml_1_2():
    # issue 155
    exec(f"""import {".".join(__name__.split(".")[:-2])}.yaml;ruamel = {".".join(__name__.split(".")[:-2])}""")
    from ruamel.yaml.compat import StringIO

    if sys.version_info >= NO_CLIB_VER:
        return
    assert ruamel.yaml.__with_libyaml__
    yaml = ruamel.yaml.YAML(typ='safe')
    yaml.version = (1, 2)
    yaml.default_flow_style = False
    data = {'a': 1, 'b': 2}
    exp = dedent("""\
    %YAML 1.2
    ---
    a: 1
    b: 2
    """)
    buf = StringIO()
    yaml.dump(data, buf)
    assert buf.getvalue() == exp



# coding: utf-8

"""
various test cases for string scalars in YAML files
'|' for preserved newlines
'>' for folded (newlines become spaces)

and the chomping modifiers:
'-' for stripping: final line break and any trailing empty lines are excluded
'+' for keeping: final line break and empty lines are preserved
''  for clipping: final line break preserved, empty lines at end not
    included in content (no modifier)

"""

import pytest
import platform

# from .compat import ordereddict
from roundtrip import round_trip, dedent, round_trip_load, round_trip_dump  # NOQA


class TestLiteralScalarString:
    def test_basic_string(self):
        round_trip("""
        a: abcdefg
        """)

    def test_quoted_integer_string(self):
        round_trip("""
        a: '12345'
        """)

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_preserve_string(self):
        inp = """
        a: |
          abc
          def
        """
        round_trip(inp, intermediate=dict(a='abc\ndef\n'))

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_preserve_string_strip(self):
        s = """
        a: |-
          abc
          def

        """
        round_trip(s, intermediate=dict(a='abc\ndef'))

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_preserve_string_keep(self):
        # with pytest.raises(AssertionError) as excinfo:
        inp = """
            a: |+
              ghi
              jkl


            b: x
            """
        round_trip(inp, intermediate=dict(a='ghi\njkl\n\n\n', b='x'))

    @pytest.mark.skipif(
        platform.python_implementation() == 'Jython', reason='Jython throws RepresenterError'
    )
    def test_preserve_string_keep_at_end(self):
        # at EOF you have to specify the ... to get proper "closure"
        # of the multiline scalar
        inp = """
            a: |+
              ghi
              jkl

            ...
            """
        round_trip(inp, intermediate=dict(a='ghi\njkl\n\n'))

    def test_fold_string(self):
        inp = """
        a: >
          abc
          def

        """
        round_trip(inp)

    def test_fold_string_strip(self):
        inp = """
        a: >-
          abc
          def

        """
        round_trip(inp)

    def test_fold_string_keep(self):
        with pytest.raises(AssertionError) as excinfo:  # NOQA
            inp = """
            a: >+
              abc
              def

            """
            round_trip(inp, intermediate=dict(a='abc def\n\n'))


class TestQuotedScalarString:
    def test_single_quoted_string(self):
        inp = """
        a: 'abc'
        """
        round_trip(inp, preserve_quotes=True)

    def test_double_quoted_string(self):
        inp = """
        a: "abc"
        """
        round_trip(inp, preserve_quotes=True)

    def test_non_preserved_double_quoted_string(self):
        inp = """
        a: "abc"
        """
        exp = """
        a: abc
        """
        round_trip(inp, outp=exp)


class TestReplace:
    """inspired by issue 110 from sandres23"""

    def test_replace_preserved_scalar_string(self):
        exec(f"""import {".".join(__name__.split(".")[:-2])};ruamel = {".".join(__name__.split(".")[:-2])}""")

        s = dedent("""\
        foo: |
          foo
          foo
          bar
          foo
        """)
        data = round_trip_load(s, preserve_quotes=True)
        so = data['foo'].replace('foo', 'bar', 2)
        assert isinstance(so, ruamel.yaml.scalarstring.LiteralScalarString)
        assert so == dedent("""
        bar
        bar
        bar
        foo
        """)

    def test_replace_double_quoted_scalar_string(self):
        exec(f"""import {".".join(__name__.split(".")[:-2])};ruamel = {".".join(__name__.split(".")[:-2])}""")

        s = dedent("""\
        foo: "foo foo bar foo"
        """)
        data = round_trip_load(s, preserve_quotes=True)
        so = data['foo'].replace('foo', 'bar', 2)
        assert isinstance(so, ruamel.yaml.scalarstring.DoubleQuotedScalarString)
        assert so == 'bar bar bar foo'


class TestWalkTree:
    def test_basic(self):
        from .comments import CommentedMap
        from .scalarstring import walk_tree

        data = CommentedMap()
        data[1] = 'a'
        data[2] = 'with\nnewline\n'
        walk_tree(data)
        exp = """\
        1: a
        2: |
          with
          newline
        """
        assert round_trip_dump(data) == dedent(exp)

    def test_map(self):
        from .compat import ordereddict
        from .comments import CommentedMap
        from .scalarstring import walk_tree, preserve_literal
        from .scalarstring import DoubleQuotedScalarString as dq
        from .scalarstring import SingleQuotedScalarString as sq

        data = CommentedMap()
        data[1] = 'a'
        data[2] = 'with\nnew : line\n'
        data[3] = '${abc}'
        data[4] = 'almost:mapping'
        m = ordereddict([('\n', preserve_literal), ('${', sq), (':', dq)])
        walk_tree(data, map=m)
        exp = """\
        1: a
        2: |
          with
          new : line
        3: '${abc}'
        4: "almost:mapping"
        """
        assert round_trip_dump(data) == dedent(exp)



import json_fix

object_lookup = {}

class Options:
    def __init__(self, data=None, default=None, _auto_generated=False, _parent_callbacks=None):
        self.default  = default or (lambda key, self_, *args: Object(
            Options(_auto_generated=True,_parent_callbacks=[self_, key])
        ))
        self.data            = data
        self._auto_generated = _auto_generated
        self._parent_callbacks = _parent_callbacks or []

    
class Object:
    def __init__(self, options_or_dict=None, **kwargs):
        this = id(self)
        options = options_or_dict if isinstance(options_or_dict, Options         ) else Options()
        a_dict  = options_or_dict if isinstance(options_or_dict, (dict, Object)) else {}
        a_dict  = to_dict(a_dict)
        a_dict.update(kwargs)
        
        self.__class__ = ObjectClass
        object_lookup[this] = (tuple(), a_dict, options.default, {}, options._parent_callbacks, options._auto_generated, len(a_dict) == 0)

class ObjectClass(Object):
    # this is "more powerful" than __getattr__
    def __getattribute__(self, key):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        if key in data:
            return data[key]
        for each_parent in lineage:
            if key in each_parent:
                return each_parent[key]
            if hasattr(each_parent, key):
                return getattr(each_parent, key)
        try:
            return object.__getattribute__(self, key)
        except Exception as error:
            return self[key]
    
    def __setattr__(self, key, value):
        this = id(self)
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[this]
        if untouched:
            if len(parent_callbacks):
                for each_parent, each_key in parent_callbacks:
                    each_parent[each_key] = self
                    parent_id = id(each_parent)
                    parent_lineage, parent_data, parent_default, parent_uninitilized_children, parent_parent_callbacks, parent_was_autogenerated, parent_untouched = object_lookup[parent_id]
                    if each_key in parent_uninitilized_children:
                        del parent_uninitilized_children[each_key]
                    object_lookup[parent_id] = (parent_lineage, parent_data, parent_default, parent_uninitilized_children, parent_parent_callbacks, parent_was_autogenerated, False)
            # touched
            object_lookup[this] = (lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, False)
        
        data[key] = value
    
    def __setitem__(self, key, value):
        this = id(self)
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[this]
        if untouched:
            if len(parent_callbacks):
                for each_parent, each_key in parent_callbacks:
                    each_parent[each_key] = self
                    parent_id = id(each_parent)
                    parent_lineage, parent_data, parent_default, parent_uninitilized_children, parent_parent_callbacks, parent_was_autogenerated, parent_untouched = object_lookup[parent_id]
                    if each_key in parent_uninitilized_children:
                        del parent_uninitilized_children[each_key]
                    object_lookup[parent_id] = (parent_lineage, parent_data, parent_default, parent_uninitilized_children, parent_parent_callbacks, parent_was_autogenerated, False)
            # touched
            object_lookup[this] = (lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, False)
        
        data[key] = value
    
    def __getattr__(self, key):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        if key in data:
            return data[key]
        for each_parent in lineage:
            if key in each_parent:
                return each_parent[key]
            if hasattr(each_parent, key):
                return getattr(each_parent, key)
        
        if key not in uninitilized_children:
            uninitilized_children[key] = default_function(key, self)
        return uninitilized_children[key]
    
    def __getitem__(self, key):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        if key in data:
            return data[key]
        for each_parent in lineage:
            if key in each_parent:
                return each_parent[key]
            if hasattr(each_parent, key):
                return getattr(each_parent, key)
        
        if key not in uninitilized_children:
            uninitilized_children[key] = default_function(key, self)
        return uninitilized_children[key]
            
    def __len__(self):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        return len(data)
    
    def __contains__(self, key):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        return key in data
    
    def __delattr__(self, key):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        
        if key in data:
            del data[key]
        if key in uninitilized_children:
            child_id = id(uninitilized_children[key])
            # detach self from the UninitilizedChild
            parent_callbacks_for[child_id] = [
                (each_parent, each_key)
                    for each_parent, each_key in parent_callbacks_for[child_id]
                    if each_key != key
            ]
            del uninitilized_children[key]
    
    def __delitem__(self, key):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        
        if key in data:
            del data[key]
        if key in uninitilized_children:
            child_id = id(uninitilized_children[key])
            # detach self from the UninitilizedChild
            parent_callbacks_for[child_id] = [
                (each_parent, each_key)
                    for each_parent, each_key in parent_callbacks_for[child_id]
                    if each_key != key
            ]
            del uninitilized_children[key]
        
    # the truthy value of obj
    def __nonzero__(self):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        if was_autogenerated and uninitilized_children:
            return False
        else:
            return True
    
    def __iter__(self):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        return data.items()
    
    def __reversed__(self):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        return reversed(data.items())
    
    def __repr__(self):
        return self.__str__()
    
    def __str__(self):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        if len(data) == 0:
            return "{}"
        return _stringify(data)
    
    def __eq__(self, other):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        return data == other
    
    def __add__(self, other):
        # this is what makes += work
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        
        if untouched and was_autogenerated:
            for each_parent, each_key in parent_callbacks:
                each_parent[each_key] = other
                return other
        else:
            if isinstance(other, dict):
                data.update(other)
                return self
            elif isinstance(other, Object):
                data.update(to_dict(other))
                return self
            else:
                # TODO: should probably be an error
                pass
    
    def __json__(self):
        lineage, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[id(self)]
        return data


def lineage(obj):
    this = id(obj)
    if this in object_lookup:
        lineage_, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[this]
        return lineage_
    else:
        return tuple()

def add_ancestor(obj, *, ancestor):
    this = id(obj)
    if this in object_lookup:
        lineage_, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[this]
        object_lookup[this] = ((*lineage_, ancestor), data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched)
    return obj

def to_dict(obj):
    this = id(obj)
    if this in object_lookup:
        lineage_, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[this]
        return data
    if isinstance(obj, dict):
        return obj
    if isinstance(obj, (tuple, list)):
        return {
            index: each
                for index, each in enumerate(obj)
        }
    if isinstance(obj, (set, frozenset)):
        return {
            each: True
                for each in obj
        }

def length(obj):
    return len(to_dict(obj))

def size(obj):
    return len(to_dict(obj))

def keys(obj):
    return list(to_dict(obj).keys())

def values(obj):
    return list(to_dict(obj).values())

def items(obj):
    return list(to_dict(obj).items())

def overwrite(obj, *args):
    data = to_dict(obj)
    for each in args:
        data.update(to_dict(each))
    return obj

def merge(*args):
    return overwrite(Object(), *args)

def copy(obj):
    original_id = id(obj)
    new_obj     = Object()
    new_id      = id(new_obj)
    
    lineage_, data, default_function, uninitilized_children, parent_callbacks, was_autogenerated, untouched = object_lookup[original_id]
    object_lookup[new_id] = (lineage_, dict(data), default_function, dict(uninitilized_children), list(parent_callbacks), False, False)
    
    return new_obj

def clear(obj):
    a_dict = to_dict(obj)
    a_dict.clear()
    return obj

def sort_keys(obj):
    a_dict = to_dict(obj)
    keys = sorted(list(a_dict.keys()))
    dict_copy = {}
    # save copy and remove
    for each_key in keys:
        dict_copy[each_key] = a_dict[each_key]
    a_dict.clear()
    # re-add in correct order
    for each_key in keys:
        a_dict[each_key] = dict_copy[each_key]
    return obj
    
# TODO
# def sort_values(obj):
#     pass

# TODO
# def grab(key_list, default):
#     pass

# TODO
# def shove(key_list, value):
#     pass

# TODO
# def recursive_merge(obj):
#     pass

Object.lineage      = lineage
Object.add_ancestor = add_ancestor
Object.to_dict      = to_dict
Object.length       = length
Object.size         = size
Object.keys         = keys
Object.values       = values
Object.items        = items
Object.overwrite    = overwrite
Object.merge        = merge
Object.copy         = copy
Object.clear        = clear
Object.sort_keys    = sort_keys

def _indent(string, by):
    indent_string = (" "*by)
    return indent_string + string.replace("\n", "\n"+indent_string)

def _stringify(value):
    onelineify_threshold = 50 # characters (of inner content)
    length = 0
    if isinstance(value, str):
        return f'"{value}"'
    elif isinstance(value, Object):
        if len(value) == 0:
            return "{}"
        items = value if isinstance(value, Object) else value.items()
        output = "{\n"
        for each_key, each_value in items:
            element_string = _stringify(each_key) + ": " + _stringify(each_value)
            length += len(element_string)+2
            output += _indent(element_string, by=4) + ", \n"
        output += "}"
        if length < onelineify_threshold:
            output = output.replace("\n    ","").replace("\n","")
        return output
    elif isinstance(value, dict):
        if len(value) == 0:
            return "{}"
        items = value if isinstance(value, Object) else value.items()
        output = "{\n"
        for each_key, each_value in items:
            element_string = _stringify(each_key) + ": " + _stringify(each_value)
            length += len(element_string)+2
            output += _indent(element_string, by=4) + ", \n"
        output += "}"
        if length < onelineify_threshold:
            output = output.replace("\n    ","").replace("\n","")
        return output
    elif isinstance(value, list):
        if len(value) == 0:
            return "[]"
        output = "[\n"
        for each_value in value:
            element_string = _stringify(each_value)
            length += len(element_string)+2
            output += _indent(element_string, by=4) + ", \n"
        output += "]"
        if length < onelineify_threshold:
            output = output.replace("\n    ","").replace("\n","")
        return output
    elif isinstance(value, set):
        if len(value) == 0:
            return "set([])"
        output = "set([\n"
        for each_value in value:
            element_string = _stringify(each_value)
            length += len(element_string)+2
            output += _indent(element_string, by=4) + ", \n"
        output += "])"
        if length < onelineify_threshold:
            output = output.replace("\n    ","").replace("\n","")
        return output
    elif isinstance(value, tuple):
        if len(value) == 0:
            return "tuple()"
        output = "(\n"
        for each_value in value:
            element_string = _stringify(each_value)
            length += len(element_string)+2
            output += _indent(element_string, by=4) + ", \n"
        output += ")"
        if length < onelineify_threshold:
            output = output.replace("\n    ","").replace("\n","")
        return output
    else:
        try:
            debug_string = value.__repr__()
        except Exception as error:
            from io import StringIO
            import builtins
            string_stream = StringIO()
            builtins.print(value, file=string_stream)
            debug_string = string_stream.getvalue()
        
        # TODO: handle "<slot wrapper '__repr__' of 'object' objects>"
        if debug_string.startswith("<class") and hasattr(value, "__name__"):
            return value.__name__
        if debug_string.startswith("<function <lambda>"):
            return "(lambda)"
        if debug_string.startswith("<function") and hasattr(value, "__name__"):
            return value.__name__
        if debug_string.startswith("<module") and hasattr(value, "__name__"):
            _, *file_path, _, _ = debug_string.split(" ")[-1]
            file_path = "".join(file_path)
            return f"module(name='{value.__name__}', path='{file_path}')"
        
        space_split = debug_string.split(" ")
        if len(space_split) >= 4 and debug_string[0] == "<" and debug_string[-1] == ">":
            
            if space_split[-1].startswith("0x") and space_split[-1] == "at":
                _, *name_pieces = space_split[0]
                *parts, name = "".join(name_pieces).split(".")
                parts_str = ".".join(parts)
                return f'{name}(from="{parts_str}")'
        
        return debug_string




# coding: utf-8

import pytest  # NOQA
from roundtrip import round_trip_load, round_trip_dump


class TestNone:
    def test_dump00(self):
        data = None
        s = round_trip_dump(data)
        assert s == 'null\n...\n'
        d = round_trip_load(s)
        assert d == data

    def test_dump01(self):
        data = None
        s = round_trip_dump(data, explicit_end=True)
        assert s == 'null\n...\n'
        d = round_trip_load(s)
        assert d == data

    def test_dump02(self):
        data = None
        s = round_trip_dump(data, explicit_end=False)
        assert s == 'null\n...\n'
        d = round_trip_load(s)
        assert d == data

    def test_dump03(self):
        data = None
        s = round_trip_dump(data, explicit_start=True)
        assert s == '---\n...\n'
        d = round_trip_load(s)
        assert d == data

    def test_dump04(self):
        data = None
        s = round_trip_dump(data, explicit_start=True, explicit_end=False)
        assert s == '---\n...\n'
        d = round_trip_load(s)
        assert d == data
